#!/usr/local/bin/perl
#
# PPSERV ver 0.75 --- "Pigeon Post" Mail-list Command Interface
#           ( 1999.07.02 named by Masato MINDA <minmin@wide.ad.jp> )
# Copyright (c)1999 by Kazunori ANDO <ando@kk.iij4u.or.jp> all rights reserved.
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
################
$PPSERV_VERSION = "0.750";
################
use IO::Socket;
if ( $0 =~ /^(\S+)\/ppserv\.pl/ ){ $ORIGIN = $1 };
$LOCK_SH = 1;
$LOCK_EX = 2;
$LOCK_NB = 4;
$LOCK_UN = 8;
$locktmp = "L$$";
#################
## exit status ##
#################
$EX_OK         =  0; # successful termination
$EX__BASE      = 64; # base value for error messages
$EX_USAGE      = 64; # command line usage error
$EX_DATAERR    = 65; # data format error
$EX_NOINPUT    = 66; # cannot open input
$EX_NOUSER     = 67; # addressee unknown
$EX_NOHOST     = 68; # host name unknown
$EX_UNAVAILABLE= 69; # service unavailable
$EX_SOFTWARE   = 70; # internal software error
$EX_OSERR      = 71; # system error (e.g., can't fork)
$EX_OSFILE     = 72; # critical OS file missing
$EX_CANTCREAT  = 73; # can't create (user) output file
$EX_IOERR      = 74; # input/output error
$EX_TEMPFAIL   = 75; # temp failure; user is invited to retry
$EX_PROTOCOL   = 76; # remote error in protocol
$EX_NOPERM     = 77; # permission denied
$EX_CONFIG     = 78; # configuration error
$EX__MAX       = 78; # maximum listed value
##### main #####
&ReadSiteConfig;     # Read sitedef.ph
&ParseInput(*e);     # Read input and set %e hash
&CheckCache(*e);     # Loop check by body-checksum and Message-Id, add cache
&MIMEdecode(*e);     # MIMEdecode if GECOS is encoded
&CommandCom(*e);     # Parse Commands
#&DebugPrint(*e);
&ExeCommand(*e);     # Exec Commands
################

sub ReadSiteConfig{
    $0 = "ReadSiteConfig";
    unshift(@INC,$ORIGIN);
    chdir $ORIGIN;
    eval("require(\"sitedef.ph\")");
    if ($@) {
	&Log("$0:ERROR:$@");
	print STDERR $@;
	exit $EX_CONFIG;
    }
    return 1;
}

sub ParseInput{
    $0 = "ParseInput";
    local(*e) = @_;
    local(@header);
    $/ = "\n\n";
    $e{'Header'} = <STDIN>;
    $/ = "\n";
    $e{'size:header'} = length($e{'Header'});
    @header = split(/\n/,$e{'Header'});
    foreach (@header){
	if (/^([^\s\:]+)\:\s*(.*)/){
	    $h = $1;
	    $c = $2;
            # field-name covert to unified form
            $horg = $h;
            $h =~ tr/A-Z/a-z/;
            # message-id -> Message-id
            if ( $h =~ /^(.)(.*)/ ){
                $hc = $1;
                $hc =~ tr/a-z/A-Z/;
                $h = $hc.$2;
            }
            # Message-id -> Message-Id
            while ( $h =~ /^(.+)(\-[a-z])(.+)/ ){
                $hc = $2;
                $hc =~ tr/a-z/A-Z/;
                $h = $1.$hc.$3;
            }
            $e{"OH:$h"} = $horg;
	    if ((length($e{"H:$h"}) > 0)&&(length($c) > 0)){
		$e{"H:$h"} .= "$_\n";
	    }elsif($c =~ /\S+/){ 
		$e{"H:$h"} .= "$c\n";
	    }
	}elsif(/^(\s+.*)/){
	    if ($h ne ""){
		$e{"H:$h"} .= "$_\n";
	    }else{
		$e{'bogus'} .= "$_\n";
	    }
	}else{
	    $e{'bogus'} .= "$_\n";
	}
    }
    while(<STDIN>){
	$e{'Body'} .= $_;
	$e{'checksum:body'} = sprintf("%011d",unpack("%31C*",$e{'Body'}));
	$e{'line:body'}++;
	$e{'size:body'} += length($_);
    }
    $e{'size'} = $e{'size:header'} + $e{'size:body'};
    
    &Log("$0:FROM: $e{'H:From'}");
    &Log("$0:DATA: header:$e{'size:header'} + body:$e{'size:body'} = $e{'size'} bytes");
    
    return 1;
}

sub CheckCache{
    $0 = "CheckCache";
    local(*e) = @_;
    local(%mc,%cm);
    local($mid);
    $mid = $e{'H:Message-Id'} || $e{'H:Message-ID'};
    $mid =~ s/\r//;
    $mid =~ s/\n//;
    if (-f $Cache){
	open(CACHE,$Cache);
	while (<CACHE>){
	    if (/^(\d{11})\s+(\<[^\>]+\>)$/){
		$mc{$2} = $1;
	    }
	}
	close(CACHE);
	if ($mc{$mid}) {
	    &Log("$0:Fail: Loop check(Message-Id:$mid)");
	    print STDERR "Fail: Loop check\n";
	    &Log("$0:Done: Article rejected");
	    exit $EX_UNAVAILABLE;
	}
    }
    open(CACHE,">>$Cache");
    print CACHE "$e{'checksum:body'} $mid\n";
    close(CACHE);
    &Log("$0:PASS: $e{'checksum:body'} $mid");
    
    return 1;
}

sub CommandCom{
    $0 = "CommandCom";
    local(*e) = @_;
    &Conv2MBr($e{'H:From'},*e);
    # for Security
    foreach (keys(%e)){
	if (/^[AMUP]\:/){
	    delete $e{$_};
	}
    }
    $e{'Body'} =~ s/\\\n//g;
    foreach (split(/\n/,$e{'Body'})){
	$passwd = ""; $comm = ""; $ml = ""; $mb = ""; $dir = "";
	next if (/^\s*$/);
	if (/^\s*approve\s+(\S+)\s+(subscribe|unsubscribe|on|off|delete|passwd|get)\s+(\S+)\s+(.*)/){
	    $passwd = $1; $comm = $2; $ml = $3; $mb = $4;
	    $ml =~ tr/A-Z/a-z/;
	    $comm =~ tr/A-Z/a-z/;
	    $dir = "$ML_SETTING_ROOT/$ml";
	    if ($ml && (-d $dir)&&(-x $dir)) {
		$e{"A:$ml $passwd $comm"} .= "$mb\n";
	    }else{
		$e{'resp'} .= "ppserv> $_\n\t>> No such ML exist.\n";
	    }
	}elsif(/^\s*approve\s+(\S+)\s+(who)\s+(\S+)/){
	    $passwd = $1; $comm = $2; $ml = $3;
	    $ml =~ tr/A-Z/a-z/;
	    $comm =~ tr/A-Z/a-z/;
	    $dir = "$ML_SETTING_ROOT/$ml";
	    if ($ml && (-d $dir)&&(-x $dir)) {
		$e{"A:$ml $passwd $comm"} .= "$comm";
	    }else{
		$e{'resp'} .= "ppserv> $_\n\t>> No such ML exist.\n";
	    }
	}elsif(/^\s*approve\s+(\S+)\s+(newinfo|help|welcome|chirm)\s+(\S+)/){
	    $passwd = $1; $comm = $2; $ml = $3;
	    $ml =~ tr/A-Z/a-z/;
	    $comm =~ tr/A-Z/a-z/;
	    $dir = "$ML_SETTING_ROOT/$ml";
	    if ($ml && (-d $dir)&&(-x $dir)) {
		$e{"A:$ml $passwd $comm"} .= "$comm";
	    }else{
		$e{'resp'} .= "ppserv> $_\n\t>> No such ML exist.\n";
	    }
	    return;
	}elsif(/^\s*(get)\s+(\S+)\s+(.*)$/){
	    $comm = $1; $ml = $2; $mb = $3;
	    $ml =~ tr/A-Z/a-z/;
	    $comm =~ tr/A-Z/a-z/;
	    $dir = "$ML_SETTING_ROOT/$ml";
	    if ($ml && (-d $dir)&&(-x $dir)) {
		$e{"M:$ml $comm"} .= "$mb\n";
	    }else{
		$e{'resp'} .= "ppserv> $_\n\t>> No such ML exist.\n";
	    }
	}elsif(/^\s*(unsubscribe|off|on|index|who)\s+(\S+)/){
	    $comm = $1; $ml = $2;
	    $ml =~ tr/A-Z/a-z/;
	    $comm =~ tr/A-Z/a-z/;
	    $dir = "$ML_SETTING_ROOT/$ml";
	    if ($ml && (-d $dir) && (-x $dir)) {
		$e{"M:$ml $comm"} .= "$comm";
	    }else{
		$e{'resp'} .= "ppserv>$_\n\t>> No such ML exist.\n";
	    }
	}elsif(/^\s*(subscribe)\s+(\S+)/){
	    $comm = $1; $ml = $2;
	    $ml =~ tr/A-Z/a-z/;
	    $comm =~ tr/A-Z/a-z/;
	    $dir = "$ML_SETTING_ROOT/$ml";
	    if ($ml && (-d $dir) && (-x $dir)) {
		$e{"U:$ml $comm"} .= "$comm";
	    }else{
		$e{'resp'} .= "ppserv> $_\n\t>> No such ML exist.\n";
	    }
	}elsif(/^\s*(confirm)\s+(\S+)\s+(\S+)/){
	    $comm = $1; $passwd = $2; $ml = $3; 
	    $ml =~ tr/A-Z/a-z/;
	    $comm =~ tr/A-Z/a-z/;
	    $dir = "$ML_SETTING_ROOT/$ml";
	    if ($ml && (-d $dir)&&(-x $dir)) {
		$e{"U:$ml $comm"} .= "$passwd";
	    }else{
		$e{'resp'} .= "ppserv> $_\n\t>> No such ML exist.\n";
	    }
	}elsif(/^\s*(info)\s+(\S+)/){
	    $comm = $1; $ml = $2;
	    $ml =~ tr/A-Z/a-z/;
	    $comm =~ tr/A-Z/a-z/;
	    $dir = "$ML_SETTING_ROOT/$ml";
	    if ($ml && (-d $dir)&&(-x $dir)) {
		$e{"U:$ml $comm"} .= "$comm";
	    }else{
		$e{'resp'} .= "ppserv> $_\n\t>> No such ML exist.\n";
	    }
	}elsif(/^\s*(generate)\s+(\S+)\s+(\S+)/){
	    $comm = $1; $ml = $2; $passwd = $3;
	    $ml =~ tr/A-Z/a-z/;
	    $comm =~ tr/A-Z/a-z/;
	    $dir = "$ML_SETTING_ROOT/$ml";
	    if ($ml && (-d $dir)&&(-x $dir)) {
		$e{'resp'} .= "ppserv> $_\n\t>> The ML already exists.\n";
	    }elsif($AUTOGENERATE){
		&AutoGenerate($ml,$passwd,*e);
	    }else{
		$e{'resp'} .= "ppserv> please set \$AUTOGENERATE = 1\; in sitedef.ph.\n";
	    }
	}elsif(/^\s*(help)/){
	    $comm = $1;
	    $comm =~ tr/A-Z/a-z/;
	    $e{"P:$comm"} .= "$comm";
	}elsif(/^\s*(end)/){
	    $e{'resp'} .= "ppserv> $_\n\t>> command end.\n";
	    return;
	}else{
	    $e{'resp'} .= "ppserv> $_\n\t>> Illegal command.\n";
	}
    }
    return 1;
}

sub PrioComm{
    local(%prio) = ('unsubscribe',1,
		    'subscribe',2,
		    'off',3,
		    'on',4,
		    'who',5,
		    'passwd',0,
		    'newinfo',0,
		    'welcome',0,
		    'delete',0,
		    'notadmin',0,
		    );
    local($comma,$commb);
    if ($a =~ /^A\:(\S+)\s+(\S+)\s+(\S+)/){
	$comma = $3;
    }else{
	$comma = 'notadmin';
    }
    if ($b =~ /^A\:(\S+)\s+(\S+)\s+(\S+)/){
	$commb = $3;
    }else{
	$commb = 'notadmin';
    }
    $prio{$comma} <=> $prio{$commb};
}

sub AutoGenerate{
    $0 = "AutoGenerate";
    local($ml,$passwd,*e) = @_;
    local($ok) = 0;
    local($dir) = "$ML_SETTING_ROOT/$ml";
    local($skeldir) = "$ML_SETTING_ROOT/skel";
    local($ucml,$tmpadmin);
    local(@admin);
    if (open(ALIASES,$ALIASES)){
	while (<ALIASES>){
	    if (/^$ml\:\s*\S+/i){
		$ok++;
	    }elsif(/^$ml\-request\:\s*(.*)/i){
		$ok++;
		$tmpadmin = $1;
		@admin = split(/\,/,$tmpadmin);
		last;
	    }
	}
	close(ALIASES);
    }else{
	&Log("$0:GENE: Can't open $ALIASES");
	$e{'resp'} .= "\n*** Can't open \$ALIASES.\n";
	return;
    }
    if ($ok != 2) {
	&Log("$0:GENE: Not find $ml entry in $ALIASES");
	return;
    }else{
	# generate dir
	mkdir($dir,0755);
	if ((-d $dir)&&(-x $dir)) {
	    &Log("$0:GENE: \$dir($dir) is created.");
	    $e{'resp'} .= "\n*** \$dir for $ml is created.\n";
	}else{
	    &Log("$0:GENE: Can't create \$dir($dir).");
	    $e{'resp'} .= "\n*** Can't create \$dir for $ml.\n";
	    return;
	}
	# generate config.ph
	if (open(SC,"$skeldir/config.ph")){
	    if (open(OC,">$dir/config.ph")){
		$ucml = $ml;
		$ucml =~ tr/a-z/A-Z/;
		while (<SC>){
		    s/skel\@/$ml\@/g;
		    s/skel\-request\@/$ml\-request\@/g;
		    s/PPML\-GENUINE/$ucml/;
		    print OC $_;
		}
		close(OC);
	    }else{
		&Log("$0:GENE: Can't create $dir/config.ph.");
		$e{'resp'} .= "\n*** Can't create config.ph for $ml.\n";
		return;
	    }
	    close(SC);
	    &Log("$0:GENE: $dir/config.ph is created.");
	    $e{'resp'} .= "\n*** config.ph for $ml is created.\n";
	}else{
	    &Log("$0:GENE: Can't open $skeldir/config.ph.");
	    $e{'resp'} .= "\n*** Can't open skel/config.ph.\n";
	    return;
	}
	# generate include
	if (open(SI,"$skeldir/include")){
	    if (open(OI,">$dir/include")){
		while (<SI>){
		    s/\/skel /\/$ml /;
		    print OI $_;
		}
		close(OI);
	    }else{
		&Log("$0:GENE: Can't create $dir/include.");
		$e{'resp'} .= "\n*** Can't create include for $ml.\n";
		return;
	    }
	    close(SI);
	    &Log("$0:GENE: $dir/include is created.");
	    $e{'resp'} .= "\n*** include for $ml is created.\n";
	}else{
	    &Log("$0:GENE: Can't open $skeldir/include.");
	    $e{'resp'} .= "\n*** Can't open skel/include.\n";
	    return;
	}	    
	if (!$OPTIMIZED_DELIVERLIST){
	    # generate empty menbers.ph
	    if (open(MEM,">$dir/members.ph")){
		print MEM "\%members = \(\n";
		print MEM "\);\n";
		print MEM "1;\n";
		close(MEM);
		&Log("$0:GENE: $dir/members.ph is created.");
		$e{'resp'} .= "\n*** empty members.ph for $ml is created.\n";
	    }else{
		&Log("$0:GENE: Can't create $dir/members.ph.");
		$e{'resp'} .= "\n*** Can't create members.ph for $ml.\n";
		return;
	    }
	}
	# generate admins.ph
	if (open(ADM,">$dir/admins.ph")){
	    print ADM "\%members = \(\n";
	    foreach (@admin){
		s/(\s*)(.*)(\s*)/$2/;
		print ADM "\'$_\',\'$passwd\',\n";
	    }
	    print ADM "\);\n";
	    print ADM "1;\n";
	    close(ADM);
	    &Log("$0:GENE: $dir/admins.ph is created.");
	    $e{'resp'} .= "\n*** admins.ph for $ml is created.\n";
	}else{
	    &Log("$0:GENE: Can't create $dir/admins.ph.");
	    $e{'resp'} .= "\n*** Can't create admins.ph for $ml.\n";
	    return;
	}
    }
    return;
}

sub ExeCommand{
    $0 = "ExeCommand";
    local(*e) = @_;
    local($ml,$passwd,$comm,$addr,$gecos,$entry);
    local(@ent);
    foreach (sort PrioComm keys(%e)){
	$ml = ""; $passwd = ""; $comm = "";
	if (/^A\:(\S+)\s+(\S+)\s+(\S+)/){
	    $ml = $1; $passwd = $2; $comm = $3;
	    $line = $_;
	    &SetML($ml,*e);
	    &PPLockWait;         # Lock
	    if (&CheckAdmin($passwd,*e)){
		&Log("$0:AUTH: Auth for $ml admin $e{'poster'} success");
		$e{'resp'} .= "\n*** Auth for $ml admin $e{'poster'} success\n\n";
		if ($comm eq "subscribe"){
		    foreach (split(/\n/,$e{$_})){
			&Log("$0:subscribe: $_");
			$e{'resp'} .= "$ml> approve $comm $_\n";
			&Conv2MBG($_,*mb);
			&Log("$0:subscribe: addr $mb{'addr'} / gecos $mb{'gecos'}");
			if (&CheckAddrForm("$mb{'addr'}",*e)){
			    if (&JoinMember($mb{'addr'},$mb{'gecos'},*e)){
				$cont = "By command from $e{'poster'},\n";
				$cont .= "address \"$mb{'addr'}\" added to $ml.\n";
				$cont .= "-- $ML_ADMIN\n";
				&NoticeMail($ML_ADMIN,"Subscribe Report($ml)",$cont);
			    }
			}
		    }
		}elsif($comm eq "unsubscribe"){
		    foreach (split(/\n/,$e{$_})){
			&Log("$0:unsubscribe: $_");
			$e{'resp'} .= "$ml> approve $comm $_\n";
			&Conv2MBG($_,*mb);
			&Log("$0:unsubscribe: addr $mb{'addr'} / gecos $mb{'gecos'}");
			if (&PPDelMember($mb{'addr'},*e)){
			    $cont = "By command from $e{'poster'},\n";
			    $cont .= "address \"$mb{'addr'}\" deleted from $ml.\n";
			    $cont .= "-- $ML_ADMIN\n";
			    &NoticeMail($ML_ADMIN,"Unsubscribe Report($ml)",$cont);
			}
		    }
		}elsif($comm eq "off"){
		    foreach (split(/\n/,$e{$_})){
			$e{'resp'} .= "$ml> approve $comm $_\n";
			&Conv2MBG($_,*mb);
			&OnOffMember($mb{'addr'},$comm,*e);
		    }
		}elsif($comm eq "on"){
		    foreach (split(/\n/,$e{$_})){
			$e{'resp'} .= "$ml> approve $comm $_\n";
			&Conv2MBG($_,*mb);
			&OnOffMember($mb{'addr'},$comm,*e);
		    }
		}elsif($comm eq "newinfo"){
		    $e{'resp'} .= "$ml> approve $comm\n";
		    &SetNewInfo(*e);
		}elsif($comm eq "welcome"){
		    $e{'resp'} .= "$ml> approve $comm\n";
		    &SetWelcome(*e);
		}elsif($comm eq "passwd"){
		    $e{'resp'} .= "$ml> approve $comm\n";
		    &ChangePass($ml,$e{$line},*e);
		}elsif($comm eq "who"){
		    $e{'resp'} .= "$ml> approve $comm\n";
		    &SendMember($ml,'1',*e);
		}elsif($comm eq "help"){
		    $e{'resp'} .= "$ml> approve $comm\n";
		    &SendHelp(*e);
		}elsif($comm eq "chirm"){
		    &SendAlive(*e);
		}elsif($comm eq "get"){
		    @ent = split(/[\n\,\ ]/,$e{$line});
                    $entry = '';
		    foreach (@ent){
			if ($entry){
			    $entry .= " $_";
			}else{
			    $entry .= "$_";
			}
		    }
		    $e{'resp'} .= "$ml> $comm $entry\n";
		    if (&GetArchive(*e,$entry)){
			&Log("$0:COMM: $comm $entry");
		    }
		}elsif($comm eq "delete"){
		    @ent = split(/[\n\,\ ]/,$e{$line});
                    $entry = '';
		    foreach (@ent){
			if ($entry){
			    $entry .= " $_";
			}else{
			    $entry .= "$_";
			}
		    }
		    $e{'resp'} .= "$ml> $comm $entry\n";
		    &DelArchive(*e,$entry);
		}
	    }else{
		&Log("$0:Fail: Auth for $ml admin $e{'poster'} failed");
		$e{'resp'} .= "\n*** Auth for $ml admin $e{'poster'} failed\n\n";
	    }
	    &UnlockDone;         # Unlock
	}elsif(/^M\:(\S+)\s+(\S+)/){
	    $ml = $1; $comm = $2;
	    &SetML($ml,*e);
	    &PPLockWait;         # Lock
	    if (&AuthMember(*e)){
		&Log("$0:AUTH: Auth for $ml member $e{'poster'} success");
		$e{'resp'} .= "\n*** Auth for $ml member $e{'poster'} success\n\n";
		unless ($USERCOMMPERM{$comm}) { 
		    $e{'resp'} .= "$ml> $comm\n\t>> $comm is not permitted.\n";
		    &UnlockDone;
		    next;
		}
		if ($comm eq "unsubscribe"){
		    $e{'resp'} .= "$ml> $comm\n";
		    if (&PPDelMember($e{'poster'},*e)){
			&Log("$0:COMM: $comm $e{'poster'}");
			$cont = "By command from $e{'poster'},\n";
			$cont .= "address \"$e{'poster'}\" deleted from $ml.\n";
			$cont .= "-- $ML_ADMIN\n";
			&NoticeMail($ML_ADMIN,"Unsubscribe Report($ml)",$cont);
		    }
		}elsif($comm eq "off"){
		    $e{'resp'} .= "$ml> $comm\n";
		    if (&OnOffMember($e{'poster'},$comm,*e)){
			&Log("$0:COMM: $comm $e{'poster'}");
			$cont = "By command from $e{'poster'},\n";
			$cont .= "address \"$e{'poster'}\" off.\n";
			$cont .= "-- $ML_ADMIN\n";
			&NoticeMail($ML_ADMIN,"OFF Report($ml)",$cont);
		    }
		}elsif($comm eq "on"){
		    $e{'resp'} .= "$ml> $comm\n";
		    if (&OnOffMember($e{'poster'},$comm,*e)){
			&Log("$0:COMM: $comm $e{'poster'}");
			$cont = "By command from $e{'poster'},\n";
			$cont .= "address \"$e{'poster'}\" on.\n";
			$cont .= "-- $ML_ADMIN\n";
			&NoticeMail($ML_ADMIN,"ON Report($ml)",$cont);
		    }
		}elsif($comm eq "who"){
		    $e{'resp'} .= "$ml> $comm\n";
		    if (&SendMember($ml,'0',*e)){
			&Log("$0:COMM: $comm");
		    }
		}elsif($comm eq "index"){
		    $e{'resp'} .= "$ml> $comm\n";
		    if (&IndexMLArc($ml,*e)){
			&Log("$0:COMM: $comm");
		    }
		}elsif($comm eq "get"){
		    @ent = split(/[\n\,\ ]/,$e{$_});
                    $entry = '';
		    foreach (@ent){
			if ($entry){
			    $entry .= " $_";
			}else{
			    $entry .= "$_";
			}
		    }
		    $e{'resp'} .= "$ml> $comm $entry\n";
		    if (&GetArchive(*e,$entry)){
			&Log("$0:COMM: $comm $entry");
		    }
		}
	    }else{
		&Log("$0:Fail: Auth for $ml member $e{'poster'} failed");
		$e{'resp'} .= "\n*** Auth for $ml member $e{'poster'} failed\n\n";
	    }
	    &UnlockDone;         # Unlock
	}elsif(/^U\:(\S+)\s+(\S+)/){
	    $ml = $1; $comm = $2;
	    &SetML($ml,*e);
	    &PPLockWait;         # Lock
	    unless ($USERCOMMPERM{$comm}) { 
		$e{'resp'} .= "$ml> $comm\n\t>> $comm is not permitted.\n";
                next;
	    }
	    $e{'resp'} .= "$ml> $comm\n";
	    if ($comm eq "subscribe"){
		&ConfirmReq($ml,*e);
	    }elsif($comm eq "confirm"){
		&ConfirmOne($ml,$e{$_},*e);
	    }elsif($comm eq "info"){
		&SendInfo(*e);
	    }
	    &UnlockDone;         # Unlock
	}elsif(/^P\:(\S+)/){
	    $comm = $1;
	    unless ($USERCOMMPERM{$comm}) { 
		$e{'resp'} .= "ppserv> $comm\n\t>> $comm is not permitted.\n";
                next;
	    }
	    if ($comm eq "help"){
		$e{'resp'} .= "ppserv> $comm\n";
		&SendHelp(*e);
	    }
	}
	chdir $ORIGIN;
	undef @DELIVERLIST; 
	undef @MEMBERLIST;
	undef @ADMINLIST;
	undef $ML_ADMIN;
	undef $ML_ADDR;
        undef %USERCOMMPERM;
	$BRACKET = "PPSERV";
	$AUTOSUBSCRIBE = 0;
	delete $INC{'admins.ph'};
	delete $INC{'members.ph'};
	delete $INC{'sitedef.ph'};
        &ReadSiteConfig;
    }
    if ($e{'resp'} ne ""){
	$e{'resp'} .= "\n-- $ML_CONTROL\n";
	&NoticeMail($e{'poster'},"PPML-Command-Report",$e{'resp'});
    }
    exit $EX_OK;
}

sub ChangePass{
    $0 = "ChangePass";
    local($ml,$passwd,*e) = @_;
    $passwd =~ s/\n//;
    eval("require(\"admins.ph\")");
    if ($@) {
	&Log("$0:ERROR:$@");
	print STDERR $@;
	&UnlockDone2;
    }
    if (defined($members{$e{'poster'}})){
	$members{$e{'poster'}} = $passwd;
	if (open(ADM,">admins.ph")){
	    print ADM "\%members = \(\n";
	    foreach $kk (keys(%members)){
		$members{$kk} =~ s/\\//g;
		$members{$kk} =~ s/\'/\\'/g;
		print ADM "\'$kk\',\'$members{$kk}\',\n";
		$mc++;
	    }
	    print ADM "\);\n";
	    print ADM "1;\n";
	    close(ADM);
	    &Log("$0:FILE: $addr\'s passwd changed.");
	}else{
	    &Log("$0:FILE: Can't open admins.ph");
	    return 0;
	}
	$e{'resp'} .= "\t>> Success to change passwd \n";
    }else{
	$e{'resp'} .= "\t>> No such admin exists.\n";
	&Log("$0:AUTH: No such admin exists");
	return 0;
    }
    return 1;
}


sub SendMember{
    $0 = "SendMember";
    local($ml,$auth,*e) = @_;
    my($file)= $DELIVERLIST[0];
    eval("require(\"$file\")");
    if ($@) {
	&Log("$0:ERROR:$@");
	print STDERR $@;
        &UnlockDone2;
    }
    eval("require(\"mimer.pl\")");
    if ($@) {
	&Log("$0:ERROR:$@");
	print STDERR $@;
        &UnlockDone2;
    }
    $e{'resp'} .= "----------- Members of $ml ----------\n";
    foreach (keys(%members)){
	if ($OPTIMIZED_DELIVERLIST){
	    if ($members{$_}[0] =~ /\=\?ISO\-2022\-JP\?B\?/){
		$members{$_}[0] = &mimedecode($members{$_});
	    }
	    if ($auth){
		$e{'resp'} .= "$members{$_}[0] <$_>:";
		for ($i = 1; defined($members{$_}[$i]); $i++){
		    $m = $members{$_}[$i];
		    $e{'resp'} .= "$m,";
		}
		$e{'resp'} .= "\n";
	    }else{
		$e{'resp'} .= "$members{$_}[0]\n";
	    }
	}else{
	    if ($members{$_} =~ /\=\?ISO\-2022\-JP\?B\?/){
		$members{$_} = &mimedecode($members{$_});
	    }
	    if ($auth){
		$e{'resp'} .= "$members{$_} <$_>\n";
	    }else{
		$e{'resp'} .= "$members{$_}\n";
	    }
	}
    }
    $e{'resp'} .= "\n";
    return 1;
}

sub IndexMLArc{
    $0 = "IndexMLArc";
    local($ml,*e) = @_;
    local(@ent);
    local($size,$num);
    if ((-d "$Savedir") && (-x "$Savedir")){
	opendir(ARC,"$Savedir");
	$ent = @ent = grep(!/^\./,readdir(ARC));
	closedir(ARC);
	if ($ent == 0){
	    $e{'resp'} .= "No archive file found\n";
	    return 1;
	}
	$e{'resp'} .= "----------- Index of $ml ----------\n";
	foreach (sort {$a <=> $b} @ent){
	    $size = sprintf("%12d",(stat("$Savedir/$_"))[7]);
	    ($sec,$min,$hour,$mday,$mon,$year)
		= (localtime((stat("$Savedir/$_"))[9]))[0..5];
	    $date = sprintf("%04d.%02d.%02d %02d:%02d:%02d",
			    1900+$year,$mon+1,$mday,$hour,$min,$sec);
	    $num  = sprintf("%09d",$_);
	    $e{'resp'} .= "$num $date $size(byte)\n";
	}
    }else{
	$e{'resp'} .= "No found archive directory(perhaps no article posted)\n";
    }
    return 1;
}


sub SetNewInfo{
    $0 = "SetNewInfo";
    local(*e) = @_;
    local($cont) = $e{'Body'};
    $cont =~ s/.*approve\s+\S+\s+newinfo\s+\S+\s*\n/\n/i;
    if (open(IN,">info")){
	print IN $cont;
	close(IN);
	&Log("$0:FILE: Set new info");
	$e{'resp'} .= "\t>> set new info file\n";
    }else{
	&Log("$0:FAIL: Can't open info");
	return 0;
    }
    return 1;
}

sub SetWelcome{
    $0 = "SetWelcome";
    local(*e) = @_;
    local($cont) = $e{'Body'};
    $cont =~ s/.*approve\s+\S+\s+welcome\s+\S+\s*\n/\n/i;
    if (open(IN,">welcome")){
	print IN $cont;
	close(IN);
	&Log("$0:FILE: Set welcome");
	$e{'resp'} .= "\t>> set welcome file\n";
    }else{
	&Log("$0:FAIL: Can't open welcome");
	return 0;
    }
    return 1;
}

sub SendHelp{
    $0 = "SendHelp";
    local(*e) = @_;
    local($cont,$file);
    local(%r);
    if ($e{'admin'}){
	if (-f "help-admin"){
	    $file = "help-admin";
	}else{
	    $file = "$ORIGIN/help-admin";
	}
    }else{
	if (-f "help"){
	    $file = "help";
	}else{
	    $file = "$ORIGIN/help";
	}
    }
    if (open(IN,$file)){
	undef $/;
	$cont = <IN>;
	$/ = "\n";
	close(IN);
	$r{'H:Mime-Version'} = "1.0";
	$r{'H:Content-Type'} = "Text/Plain; charset=iso-2022-jp";
	&NoticeMail($e{'poster'},"PPML HELP",$cont,*r);
	$e{'resp'} .= "\t>> send help\n";
    }else{
	&Log("$0:FAIL: Can't open $file");
	return 0;
    }
    return 1;
}

## Just kidding
sub SendAlive{
    $0 = "SendAlive";
    local(*e) = @_;
    local($cont,$file);
    local(%r);
    if ($e{'admin'}){
	$file = "alive";
    }else{
	return 0;
    }
    open(IN ,$file);
    $cont = <IN>;
    close(IN);
    $r{'H:Mime-Version'} = "1.0";
    $r{'H:Content-Type'} = "Text/Plain; charset=iso-2022-jp";
    &NoticeMail($e{'poster'},"Pigeon Post $PPSERV_VERSION IS ALIVE",$cont,*r);
    return 1;
}

sub SendInfo{
    $0 = "SendInfo";
    local(*e) = @_;
    local($cont);
    local(%r);
    if (open(IN,"info")){
	undef $/;
	$cont = <IN>;
	$/ = "\n";
	close(IN);
	$r{'H:Mime-Version'} = "1.0";
	$r{'H:Content-Type'} = "Text/Plain; charset=iso-2022-jp";
	&NoticeMail($e{'poster'},"Info $ml",$cont,*r);
	$e{'resp'} .= "\t>> send info\n";
    }else{
	&Log("$0:FAIL: Can't open info");
	return 0;
    }
    return 1;
}

sub ConfirmOne{
    $0 = "ConfirmOne";
    local($ml,$pass,*e) = @_;
    local($gecos);
    if (-f "confirm.ph"){
	delete $INC{'confirm.ph'};
	eval("require(\"confirm.ph\")");
	if ($@) {
	    &Log("$0:ERROR:$@");
	    return 0;
	}
	if ($confirm{$e{'poster'}} eq $pass){
	    &Log("$0:PASS: $e{'poster'} pass");
	    $e{'resp'} .= "\t>> confirm:PASS: $e{'poster'}\n";
	    if ($AUTOSUBSCRIBE){
		if (&JoinMember($e{'poster'},$e{'gecos'},*e)){
		    $cont = "By command from $e{'poster'},\n";
		    $cont .= "address \"$e{'poster'}\" added to $ML_ADDR.\n";
		    $cont .= "-- $ML_ADMIN\n";
		    &NoticeMail($ML_ADMIN,"Subscribe Report($ml)",$cont);
		    $cont = "\nYour confirmation for \"$ml\" mail-list completed,";
		    $cont .= "\nand subscribe your address to $ML_ADDR.";
		    $cont .= "\n\nThanks.\n-- $ML_ADMIN";
		    &NoticeMail($e{'poster'},'Confirmation Completed',$cont);
		}else{
		    &Log("$0:FAIL: Auto subscribe failed");
		}
	    }else{
		&Log("$0:PASS: request forwarded to $ML_ADMIN");
		$e{'resp'} .= "\t>> confirm:DONE: request forwarded to $ML_ADMIN\n";
		$cont = "\nYour confirmation for \"$ml\" mail-list completed.";
		$cont .= "\nSubscribe request forwarded to $ML_ADMIN.";
		$cont .= "\nPlease wait a while\n\nThanks.\n-- $ML_ADMIN";
		&NoticeMail($e{'poster'},'Confirmation Completed',$cont);
		eval("require(\"mimer.pl\")");
		if ($@) {
		    &Log("$0:ERROR:$@");
		    print STDERR $@;
		    &UnlockDone2;
		}
		$gecos = &mimedecode($e{'gecos'});
		$cont = "\nConfirmation from $e{'poster'} for \"$ml\" mail-list\n";
		$cont .= "completed. If you approve, you can send a line below with password.";
		$cont .= "\n\n\tapprove PASSWORD subscribe $ml $gecos <$e{'poster'}>";
		$cont .= "\n\nThanks.\n-- $ML_ADMIN";
		&NoticeMail($ML_ADMIN,'Confirmation Completed',$cont);
	    }
	    $0 = "ConfirmOne";
	    ###
	    delete $confirm{$e{'poster'}};
	    if (open(MEM,">confirm.ph")){
		print MEM "\%confirm = \(\n";
		foreach $kk (keys(%confirm)){
		    print MEM "\'$kk\',\'$confirm{$kk}\',\n";
		    $mc++;
		}
		print MEM "\);\n";
		print MEM "1;\n";
		close(MEM);
		&Log("$0:FILE: $e{'poster'}($pass) completed");
	    }else{
		&Log("$0:FILE: Can't open confirm.ph");
		return 0;
	    }
	}else{
	    &Log("$0:Fail: $e{'poster'} fail.");
	    $cont = "\nYour confirmation for \"$ml\" mail-list failed.";
	    $cont .= "\nPlease retry subscribe and confirm.";
	    $cont .= "\n\nThanks.\n-- $ML_ADMIN";
	    &NoticeMail($e{'poster'},'Confirm Failed',$cont);
	    $e{'resp'} .= "\t>> Confirm failed\n";
	}
    }else{
	&Log("$0:Fail: $ml has no confirm.ph.");
	$cont = "\nYour confirmation for \"$ml\" mail-list failed.";
	$cont .= "\nPlease retry subscribe and confirm.";
	$cont .= "\n\nThanks.\n-- $ML_ADMIN";
	&NoticeMail($e{'poster'},'Confirm Failed',$cont);
	$e{'resp'} .= "\t>> Confirm failed\n";
    }
    return 1;
}

sub ConfirmReq{
    $0 = "ConfirmReq";
    local($ml,*e) = @_;
    local($pass) = time;
    $pass .= "-$e{'checksum:body'}";
    if (&AuthMember(*e)){
	&Log("$0:FILE: $e{'poster'} is already a member of $ml");
	$e{'resp'} .= "\t>> subscribe:FAIL: $e{'poster'} is already a member\n";
	return 1;
    }
    if (-f "confirm.ph"){
	delete $INC{'confirm.ph'};
	eval("require(\"confirm.ph\")");
	if ($@) {
	    &Log("$0:ERROR:$@");
	    print STDERR $@;
	    &UnlockDone2;
	}
    }
    $confirm{$e{'poster'}} = $pass;
    if (open(MEM,">confirm.ph")){
	print MEM "\%confirm = \(\n";
	foreach $kk (keys(%confirm)){
	    print MEM "\'$kk\',\'$confirm{$kk}\',\n";
	    $mc++;
        }
	print MEM "\);\n";
	print MEM "1;\n";
	close(MEM);
	&Log("$0:FILE: $e{'poster'}($pass) add to confirm.ph");
	$e{'resp'} .= "\t>> subscribe:DONE: confirm request to $e{'poster'}\n";
    }else{
	&Log("$0:FILE: Can't open confirm.ph");
	return 0;
    }
    $cont = "\nSubscribe request for \"$ml\" mail-list from your address received.";
    $cont .= "\nIf you can approve this request, Send back a line below.";
    $cont .= "\n\tconfirm $pass $ml\n\nThanks.\n-- $ML_ADMIN";
    &NoticeMail($e{'poster'},'Request to Confirm',$cont);
    return 1;
}

sub GetArchive{
    $0 = "GetArchive";
    local(*e,$entry) = @_;
    local(@ent,$sp,$ep,$sc);
    foreach (split(/\s+/,$entry)){
	if (/^(\d+)\-(\d+)$/){
	    $sp = $1; $ep = $2;
	    if (($sp > 0) && ($ep > $sp)){
		for ($i = $sp; $i <= $ep; $i++ ){
		    if ( -f "$Savedir/$i"){
			push(@ent,"$i");
		    }
		}
	    }
	}elsif(/^(\d+)$/){
	    $sp = $1;
	    if ($sp > 0){
		if (-f "$Savedir/$sp"){
		    push(@ent,"$sp");
		}
	    }
	}
    }
    &Log("$0:FILE: @ent");
    $e{'resp'} .= "\t>> get @ent accepted\n";
    foreach (@ent){
	&PPSendFile($e{'poster'},$_);
	$sc++;
	if (($sc%5) == 0){
	    sleep 5;
	}
    }
    return 1;
}

sub DelArchive{
    $0 = "DelArchive";
    local(*e,$entry) = @_;
    local(@ent,$sp,$ep);
    local($sc) = 0;
    foreach (split(/\s+/,$entry)){
	if (/^(\d+)\-(\d+)$/){
	    $sp = $1; $ep = $2;
	    if (($sp > 0) && ($ep > $sp)){
		for ($i = $sp; $i <= $ep; $i++ ){
		    if ( -f "$Savedir/$i"){
			push(@ent,"$i");
			$sc++;
		    }
		}
	    }
	}elsif(/^(\d+)$/){
	    $sp = $1;
	    if ($sp > 0){
		if (-f "$Savedir/$sp"){
		    push(@ent,"$sp");
		    $sc++;
		}
	    }
	}
    }
    if ($sc == 0) {
	&Log("$0:FILE: No delete file exist");
	$e{'resp'} .= "\t>> No delete file exist\n";
	return 0;
    }
    &Log("$0:FILE: @ent");
    $e{'resp'} .= "\t>> delete @ent accepted\n";
    foreach (@ent){
	unlink("$Savedir/$_");
    }
    return 1;
}

sub JoinMember{
    $0 = "JoinMember";
    local($addr,$gecos,*e) = @_;
    local($mc) = 0;
    local($lcaddr) = $addr;
    local($tk);
    $lcaddr = &FormatAddr($lcaddr);
    my($file) = $DELIVERLIST[0];
    eval("require(\"$file\")");
    if ($@) {
	&Log("$0:ERROR:$@");
	print STDERR $@;
        &UnlockDone2;
    }
    unless ($OPTIMIZED_DELIVERLIST){
	foreach $kk (keys(%members)){
	    $tk = $kk;
	    $tk = &FormatAddr($tk);
	    $members{$tk} = $members{$kk};
	    if ($tk ne $kk){
		delete $members{$kk};
	    }
	}
    }
    if (length($addr) == 0){
	&Log("$0:FAIL: null address.");
	return 0;
    }
    if (length($gecos) == 0){
	$gecos = "No GECOS";
    }
    ### Just add one address
    if ($OPTIMIZED_DELIVERLIST){
	$members{$lcaddr}[0] = $gecos;
	for ($i = 1; $members{$lcaddr}[$i] ne ""; $i++){
	    $m = $members{$lcaddr}[$i];
	    if ($ml eq $m){
		$e{'resp'} .= "\t>> subscribe:FAIL: $addr is already exist on members\n";
		return 0;
	    }
	}
	### add to ML
	$members{$lcaddr}[$i] = $ml;
    }else{
	if ((!defined($members{$lcaddr})) && (!defined($members{"\# $lcaddr"}))){
	    $members{$lcaddr} = $gecos;
	}else{
	    $e{'resp'} .= "\t>> subscribe:FAIL: $addr is already exist on members\n";
	    return 0;
	}
    }
    eval("require(\"mimew.pl\")");
    if ($@) {
	&Log("$0:ERROR:$@");
	print STDERR $@;
	&UnlockDone2;
    }
    if (open(MEM,"> $file")){
	print MEM "\%members = \(\n";
	foreach $kk (keys(%members)){
	    if ($OPTIMIZED_DELIVERLIST){
		if ($members{$kk}[0] =~ /\033../){
		    $members{$kk}[0] = &mimeencode($members{$kk}[0]);
		    $members{$kk}[0] =~ s/\n//g;
		}
		$members{$kk}[0] =~ s/\\//g; 
		$members{$kk}[0] =~ s/\'/\\'/g; #'
		print MEM "\'$kk\' => [\'$members{$kk}[0]\',";
		for ($i = 1; $members{$kk}[$i] ne ""; $i++){
		    $op = $members{$kk}[$i];
		    print MEM "\'$op\',";
		}
		print MEM "],\n";
		$mc++;
	    }else{
		if ($members{$kk} =~ /\033../){
		    $members{$kk} = &mimeencode($members{$kk});
		    $members{$kk} =~ s/\n//g;
		}
		$members{$kk} =~ s/\\//g; 
		$members{$kk} =~ s/\'/\\'/g; #'
		print MEM "\'$kk\',\'$members{$kk}\',\n";
		$mc++;
	    }
	}
	print MEM "\);\n";
	print MEM "1;\n";
	close(MEM);
	&Log("$0:FILE: $addr ($gecos) add to members($mc).");
	$e{'resp'} .= "\t>> subscribe:DONE: $addr add to members\n";
    }else{
	&Log("$0:FILE: Can't open $file"); #'
	return 0;
    }
    if (-f "welcome"){
	$cont  = "\nNow, one of $ml-admins add your address to the list.";
	$cont .= "\nIf you have any questions, please send your question";
	$cont .= "\nto $ML_ADMIN.";
	$cont .= "\nThe welcome message is shown below.";
	$cont .= "\nThank you.";
	$cont .= "\n-- $ML_ADMIN\n\n";
	open(WELCOME, "welcome");
	while (<WELCOME>){
	    $cont .= $_;
	}
	&NoticeMail($addr,"Your address subscribed to $ml",$cont);
    }
    return 1;
}

sub PPDelMember{
    $0 = "PPDelMember";
    local($addr,*e) = @_;
    local($mc) = 0;
    my($file) = $DELIVERLIST[0];
    eval("require(\"$file\")");
    if ($@) {
	&Log("$0:ERROR:$@");
	print STDERR $@;
	&UnlockDone2;
    }
    ### Just delete one address
    if ($OPTIMIZED_DELIVERLIST){
	if (defined($members{$addr})){
	    $gecos = $members{$addr}[0];
	    for ($i = 1; defined($members{$addr}[$i]); $i++){
		$m = $members{$addr}[$i];
		push(@mls,$m) unless ($m eq $ml);
	    }
	    delete $members{$addr};
	    $members{$addr}[0] = $gecos;
	    foreach $m (@mls){
		push(@{$members{$addr}},$m);
	    }
	}elsif(defined($members{"\# $addr"})){
	    $gecos = $members{"\# $addr"}[0];
	    for ($i = 1; defined($members{"\# $addr"}[$i]); $i++){
		$m = $members{"\# $addr"}[$i];
		push(@mls,$m) unless ($m eq $ml);
	    }
	    delete $members{"\# $addr"};
	    $members{"\# $addr"}[0] = $gecos;
	    foreach $m (@mls){
		push(@{$members{"\# $addr"}},$m);
	    }
	}else{
	    &Log("$0:FAIL: $addr is not members");
	    $e{'resp'} .= "\t>> unsubscribe:FAIL: $addr is not members\n";
	    return 0;
	}
    }else{
	if (defined($members{$addr}) || defined($members{"\# $addr"})){
	    delete $members{$addr};
	    delete $members{"\# $addr"};
	}else{
	    &Log("$0:FAIL: $addr is not members");
	    $e{'resp'} .= "\t>> unsubscribe:FAIL: $addr is not members\n";
	    return 0;
	}
    }
    if (open(MEM,"> $file")){
	print MEM "\%members = \(\n";
	foreach $kk (keys(%members)){
	    if ($OPTIMIZED_DELIVERLIST){
		if ($members{$kk}[0] =~ /\033../){
		    $members{$kk}[0] = &mimeencode($members{$kk}[0]);
		    $members{$kk}[0] =~ s/\n//g;
		}
		$members{$kk}[0] =~ s/\\//g; 
		$members{$kk}[0] =~ s/\'/\\'/g; #'
		print MEM "\'$kk\' => [\'$members{$kk}[0]\',";
		for ($i = 1; defined($members{$kk}[$i]); $i++){
		    $op = $members{$kk}[$i];
		    print MEM "\'$op\',";
		}
		print MEM "],\n";
		$mc++;
	    }else{
		if ($members{$kk} =~ /\033../){
		    $members{$kk} = &mimeencode($members{$kk});
		    $members{$kk} =~ s/\n//g;
		}
		$members{$kk} =~ s/\\//g; 
		$members{$kk} =~ s/\'/\\'/g; #'
		print MEM "\'$kk\',\'$members{$kk}\',\n";
		$mc++;
	    }
	}
	print MEM "\);\n";
	print MEM "1;\n";
	close(MEM);
	&Log("$0:FILE: $addr delete from members($mc).");
	$e{'resp'} .= "\t>> unsubscribe:DONE: $addr is deleted from members\n";
    }else{
	&Log("$0:FILE: Can't open $file");
	return 0;
    }
    return 1;
}

sub OnOffMember{
    $0 = "OnOffMember";
    local($addr,$comm,*e) = @_;
    local($mc) = 0;
    my($file) = $DELIVERLIST[0];
    eval("require(\"$file\")");
    if ($@) {
	&Log("$0:ERROR:$@");
	print STDERR $@;
        &UnlockDone2;
    }
    if ($comm eq 'off'){
	if ($OPTIMIZED_DELIVERLIST){
	    ### Just off one address
	    my($change)=0;
	    for ($i = 1; defined($members{$addr}[$i]); $i++){
		if ($members{$addr}[$i] eq $ml){
		    $members{"$addr"}[$i] = "\# $ml";
		    $change++;
		}
	    }
	    if ($change == 0){
		$e{'resp'} .= "\t>> off:FAIL: Not active $addr on members\n";
		return 0;
	    }
	}else{
	    ### Just off one address
	    if ($members{$addr}){
		$members{"\# $addr"} = $members{$addr};
		delete $members{$addr};
	    }else{
		$e{'resp'} .= "\t>> off:FAIL: Not active $addr on members\n";
		return 0;
	    }
	}
    }else{
	if ($OPTIMIZED_DELIVERLIST){
	    ### Just on one address
	    my($change)=0;
	    for ($i = 1; defined($members{$addr}[$i]); $i++){
		if ($members{$addr}[$i] eq "\# $ml"){
		    $members{"$addr"}[$i] = "$ml";
		    $change++;
		}
	    }
	    if ($change == 0){
		$e{'resp'} .= "\t>> on:FAIL: Not active $addr on members\n";
		return 0;
	    }
	}else{
	    ### Just on one address
	    if ($members{"\# $addr"}){
		$members{$addr} = $members{"\# $addr"};
		delete $members{"\# $addr"};
	    }else{
		$e{'resp'} .= "\t>> on:FAIL: $addr is not found in members\n";
		return 0;
	    }
	}
    }
    if (open(MEM,"> $file")){
	print MEM "\%members = \(\n";
	foreach $kk (keys(%members)){
	    if ($OPTIMIZED_DELIVERLIST){
		if ($members{$kk}[0] =~ /\033../){
		    $members{$kk}[0] = &mimeencode($members{$kk}[0]);
		    $members{$kk}[0] =~ s/\n//g;
		}
		$members{$kk}[0] =~ s/\\//g; 
		$members{$kk}[0] =~ s/\'/\\'/g; #'
		print MEM "\'$kk\' => [\'$members{$kk}[0]\',";
		for ($i = 1; defined($members{$kk}[$i]); $i++){
		    $op = $members{$kk}[$i];
		    print MEM "\'$op\',";
		}
		print MEM "],\n";
		$mc++;
	    }else{
		if ($members{$kk} =~ /\033../){
		    $members{$kk} = &mimeencode($members{$kk});
		    $members{$kk} =~ s/\n//g;
		}
		$members{$kk} =~ s/\\//g; 
		$members{$kk} =~ s/\'/\\'/g; #'
		print MEM "\'$kk\',\'$members{$kk}\',\n";
		$mc++;
	    }
        }
	print MEM "\);\n";
	print MEM "1;\n";
	close(MEM);
	&Log("$0:FILE: $addr $comm in members($mc).");
	$e{'resp'} .= "\t>> $comm:DONE: $comm $addr in members\n";
    }else{
	&Log("$0:FILE: Can't open $file");
	return 0;
    }
    return 1;
}

sub Conv2MBr{
    local($ent,*e) = @_;
    local($na);
    $ent =~ s/\n//g;
    eval("require(\"mimew.pl\")");
    if ($@) {
	&Log("$0:ERROR:$@");
	print STDERR $@;
        exit $EX_CONFIG;
    }
    if ($ent =~ /\033../){
	$ent = &mimeencode($ent);
	$ent =~ s/\n//g;
    }
    if ($ent =~ /^\s*([^\<\>]*)\s*\<(\S+)\>[\s\,]*/){
	$na++; $e{'gecos'} = $1; $e{'poster'} = $2; 
    }elsif($ent =~ /^\s*([^\(\)\s]+)\s*\(([^\)]*)\)[\s\,]*/){
	$na++; $e{'gecos'} = $2; $e{'poster'} = $1; 
    }elsif($ent =~ /^\s*([^\(\)\<\>\s]+)[\s\,]*/){
	$na++; $e{'poster'} = $1; 
    }
    if ($na == 0) { return 0; };
    $e{'gecos'} =~ s/\"//g; #"
    $e{'gecos'} =~ s/^(.*)\s+$/$1/;
    $e{'gecos'} =~ s/^\s+(.*)$/$1/;
    $e{'poster'} =  &FormatAddr($e{'poster'});
    return 1;
}

sub Conv2MBG{
    local($ent,*mb) = @_;
    $mb{'addr'} = ""; $mb{'gecos'} = "";
    $ent =~ s/\n//g;
    eval("require(\"mimew.pl\")");
    if ($@) {
	&Log("$0:ERROR:$@");
	print STDERR $@;
        &UnlockDone2;
    }
    if ($ent =~ /\033../){
	$ent = &mimeencode($ent);
	$ent =~ s/\n//g;
    }
    if ($ent =~ /^\s*([^\<\>]*)\s*\<(\S+)\>\s*$/){
	$mb{'gecos'} = $1; $mb{'addr'} = $2; 
    }elsif($ent =~ /^\s*([^\(\)\s]+)\s*\(([^\)]*)\)\s*$/){
	$mb{'gecos'} = $2; $mb{'addr'} = $1; 
    }elsif($ent =~ /^\s*([^\(\)\<\>\s]+)\s*$/){
	$mb{'addr'} = $1; 
    }
    $mb{'gecos'} =~ s/\"//g; #"
    $mb{'gecos'} =~ s/^(.*)\s+$/$1/;
    $mb{'gecos'} =~ s/^\s+(.*)$/$1/;
    $mb{'addr'} = &FormatAddr($mb{'addr'});
    return 1;
}

sub AuthMember{
    $0 = "AuthMember";
    local(*e) = @_;
    local($addr) = $e{'poster'};
    $addr =  &FormatAddr($addr);
    $lcaddr = &LcAddr($addr);
    foreach $list (@ADMINLIST,@DELIVERLIST){
	delete $INC{$list};
	eval("require(\"$list\")");
	if ($@) {
	    &Log("$0:ERROR:$@");
	    print STDERR $@;
	    &UnlockDone2;
	}
	if ($OPTIMIZED_DELIVERLIST == 0){
	    if (defined($members{$addr})){
		&Log("$0:AUTH: $addr found in $list.");
		return 1;
	    }elsif(defined($members{"\# $addr"})){
		&Log("$0:AUTH: $addr found in $list(off).");
		return 1;
	    }elsif(defined($members{"$lcaddr"})){
		&Log("$0:AUTH: $addr found in $list(old format).");
		return 1;
	    }elsif(defined($members{"\# $lcaddr"})){
		&Log("$0:AUTH: $addr found in $list(off/old format).");
		return 1;
	    }
	}else{
	    if ($members{$addr}[1] ne "" ){
		for ($i = 1; $members{$addr}[$i] ne ""; $i++){
		    $m = $members{$addr}[$i];
		    if ($m eq $ml){
			$e{'poster'} = $addr;
			&Log("$0:AUTH: $addr found in $list.");
			return 1;
		    }elsif($m eq "# $ml"){
			$e{'poster'} = $addr;
			&Log("$0:AUTH: $addr found in $list(off).");
			return 1;
		    }
		}
	    }elsif($members{$lcaddr}[1] ne "" ){
		for ($i = 1; $members{$lcaddr}[$i] ne ""; $i++){
		    $m = $members{$lcaddr}[$i];
		    if ($m eq $ml){
			$e{'poster'} = $addr;
			&Log("$0:AUTH: $addr found in $list(old format).");
			return 1;
		    }elsif($m eq "# $ml"){
			$e{'poster'} = $addr;
			&Log("$0:AUTH: $addr found in $list(off/old format).");
			return 1;
		    }
		}
	    }else{
		if (defined($members{$addr})){
		    &Log("$0:AUTH: $addr found in $list.");
		    return 1;
		}elsif(defined($members{"\# $addr"})){
		    &Log("$0:AUTH: $addr found in $list(off).");
		    return 1;
		}elsif(defined($members{"$lcaddr"})){
		    &Log("$0:AUTH: $addr found in $list(old format).");
		    return 1;
		}elsif(defined($members{"\# $lcaddr"})){
		    &Log("$0:AUTH: $addr found in $list(off/old format).");
		    return 1;
		}
	    }
	}
    }
    &Log("$0:Fail: Not in members : $addr");
    return 0;
}    

sub CheckAdmin{
    $0 = "CheckAdmin";
    local($passwd,*e) = @_;
    local($addr) = $e{'poster'};
    local($lcaddr) = &LcAddr($addr);
    &Log("$0:FILE: @ADMINLIST / $addr");
    foreach $list (@ADMINLIST){
	delete $INC{$list};
	eval("require(\"$list\")");
	if ($@) {
	    &Log("$0:ERROR:$@");
	    print STDERR $@;
	    &UnlockDone2;
	}
	foreach $mem (keys(%members)){
	    $mem = &FormatAddr($mem);
	    &Log("$0:Try: $mem / $addr") if ($debug > 8);
	    if (($addr eq $mem) && ($passwd eq $members{$addr})){
		&Log("$0:AUTH: $addr found in $list and passwd matched.");
		$e{'admin'} = 1;
		return 1;
	    }elsif(($lcaddr eq $mem) && ($passwd eq $members{$lcaddr})){
		&Log("$0:AUTH: $addr found in $list(old format) and passwd matched.");
		$e{'admin'} = 1;
		return 1;
	    }
	    &Log("$0:Fail: $mem / $addr") if ($debug > 8);
	}
    }
    &Log("$0:Fail: Not admin");
    return 0;
}    

sub MIMEdecode{
    $0 = "MIMEdecode";
    local(*e) = @_;
    local($dc) = 0;
    eval("require(\"mimer.pl\")");
    if ($@) {
	&Log("$0:ERROR:$@");
	print STDERR $@;
	exit $EX_CONFIG;
    }
    foreach $hh (keys(%e)){
	if ( $hh =~ /^H\:/ ){
	    if ( $e{$hh} =~ /\=\?ISO\-2022\-JP\?B\?/i ){
		$e{$hh} = &mimedecode($e{$hh});
		$dc++;
	    }
	}
    }
    &Log("$0:MIME: Decode $dc headers.");
    return 1;
}

sub MIMEencode{
    $0 = "MIMEencode";
    local(*e) = @_;
    local($ec)= 0;
    eval("require(\"mimew.pl\")");
    if ($@) {
	&Log("$0:ERROR:$@");
	print STDERR $@;
	exit $EX_CONFIG;
    }
    foreach $hh (keys(%e)){
	if ( $hh =~ /^H\:/ ){
	    if ( $e{$hh} =~ /\033../ ){
		$e{$hh} = &mimeencode($e{$hh});
		$ec++;
	    }
	}
    }
    &Log("$0:MIME: Encode $ec headers.");
    return 1;
}

sub NoticeMail{
    $0 = "NoticeMail";
    local($rcpt,$sub,$cont,*r) = @_;
    local($rem);
    local($rc) = 0;
    ####
    $r{'H:Subject'} = $sub;
    $r{'H:From'} = "PPML-Server <$ML_CONTROL>";
    $r{'H:To'} = $rcpt;
    $ML_SERVER_ADMIN = $ML_SERVER_ADMIN || $ML_CONTROL;
    #### SENDMAIL
    if ($SENDMAIL){
	if (open(SC,"|$SENDMAIL -f $ML_SERVER_ADMIN -- $rcpt")){
	    foreach $hh (keys(%r)){
		if ($hh =~ /^H\:(\S+)$/){
		    print SC "$1\: $r{$hh}\r\n";
		}
	    }
	    print SC "\n";
	    print SC "$cont\n";
	    close(SC);
	    &Log("$0:PIPE: Put notice to $SENDMAIL");
	}else{
	    &Log("$0:Fail: Can't put notice to $SENDMAIL");
	    &UnlockDone2;
	}
	return 1;
    }
    #### SMTP
    $cont =~ s/\n\./\n\.\./g;
    while (!$rem){
	$rem = IO::Socket::INET->new(Proto =>"tcp",PeerAddr => $SMTP_HOST,);
	sleep 3;
    }
    $rem->autoflush(1);
    $resp = <$rem>; &Log("$0:SMTP: $resp") if ($debug > 1);
    print $rem "HELO $PPML_HOST\r\n";
    $resp = <$rem>; &Log("$0:SMTP: $resp") if ($debug > 1);
    print $rem "Mail From:$ML_SERVER_ADMIN\r\n";
    $resp = <$rem>; &Log("$0:SMTP: $resp") if ($debug > 1);
    print $rem "Rcpt To:<$rcpt>\r\n";
    $resp = <$rem>; 
    if ($resp =~ /^250/){
	$rc++;
	&Log("$0:SMTP: $resp") if ($debug > 1);
    }else{
	&Log("$0:SMTP: $resp");
    }
    if ($rc > 0){
	&Log("$0:SMTP: $BRACKET $e{'seq'} : $rc rcpts OK.");
    }else{
	&Log("$0:Fail: $BRACKET $e{'seq'} : $rc rcpts.");
        &UnlockDone2;
    }
    print $rem "Data\r\n";
    $resp = <$rem>; &Log("$0:SMTP: $resp") if ($debug > 1);
    foreach $hh (keys(%r)){
	if ($hh =~ /^H\:(\S+)$/){
	    print $rem "$1\: $r{$hh}\r\n";
	}
    }
    print $rem "\r\n";
    $cont =~ s/([^\r])\n/$1\r\n/g;
    print $rem "$cont\r\n.\r\n";
    $resp = <$rem>; &Log("$0:SMTP: $resp");
    print $rem "QUIT\r\n";
    $resp = <$rem>; &Log("$0:SMTP: $resp");
    return  1;
}

sub PPSendFile{
    $0 = "PPSendFile";
    local($rcpt,$art) = @_;
    local($rem);
    local($rc) = 0;
    ####
    $ML_SERVER_ADMIN = $ML_SERVER_ADMIN || $ML_CONTROL;
    ####
    if (open(AM,"$Savedir/$art")){
	undef $/;
	$cont = <AM>;
	$/ = "\n";
	close(AM);
    }else{
	&Log("$0:FAIL: Can't open $Savedir/$art");
	return 0;
    }
    #### SENDMAIL
    if ($SENDMAIL){
	if (open(SC,"|$SENDMAIL -f $ML_ADMIN -- $rcpt")){
	    print SC "$cont\n";
	    close(SC);
	    &Log("$0:PIPE: Put archive to $SENDMAIL");
	}else{
	    &Log("$0:Fail: Can not put archive to $SENDMAIL");
	    return 0;
	}
	return 1;
    }
    #### SMTP
    $cont =~ s/\n\./\n\.\./g;
    while (!$rem){
	$rem = IO::Socket::INET->new(Proto =>"tcp",PeerAddr => $SMTP_HOST,);
	sleep 3;
    }
    $rem->autoflush(1);
    $resp = <$rem>; &Log("$0:SMTP: $resp") if ($debug > 1);
    print $rem "HELO $PPML_HOST\r\n";
    $resp = <$rem>; &Log("$0:SMTP: $resp") if ($debug > 1);
    print $rem "Mail From:$ML_SERVER_ADMIN\r\n";
    $resp = <$rem>; &Log("$0:SMTP: $resp") if ($debug > 1);
    print $rem "Rcpt To:<$rcpt>\r\n";
    $resp = <$rem>; 
    if ($resp =~ /^250/){
	$rc++;
	&Log("$0:SMTP: $resp") if ($debug > 1);
    }else{
	&Log("$0:SMTP: $resp");
    }
    if ($rc > 0){
	&Log("$0:SMTP: $BRACKET $art : $rc rcpts OK.");
    }else{
	&Log("$0:Fail: $BRACKET $art : $rc rcpts.");
        &UnlockDone2;
    }
    print $rem "Data\r\n";
    $resp = <$rem>; &Log("$0:SMTP: $resp") if ($debug > 1);
    $cont =~ s/([^\r])\n/$1\r\n/g;
    print $rem "$cont\r\n.\r\n";
    $resp = <$rem>; &Log("$0:SMTP: $resp");
    print $rem "QUIT\r\n";
    $resp = <$rem>; &Log("$0:SMTP: $resp");
    return  1;
}

sub ChdirByArg{
    $0 = "ChdirByArg";
    if ((-d $DIR) && (-x $DIR)){
	unshift(@INC, $DIR);
	chdir $DIR;
    }else{
	print STDERR "$0:WARN: \$DIR is not given or Can't chdir.";
	exit $EX_NOPERM;
    }
    return 1;
}

sub CheckAddrForm
{
    $0 = "CheckAddrForm";
    local($addr,*e) = @_;
    local($up,$dom);
    $addr = &FormatAddr($addr);
    $addr =~ s/^\s*(\S+)\s*$/$1/;
    &Log("$0:Check:$addr");
    $num = ($up,$dom) = split(/\@/,$addr);
    if ( $up =~ /^[^,:;\x00-\x20\<\>\(\)\[\]\\\@\"\x7f]+$/ ){ #"]/){
        # OK
    }else{
	$e{'resp'} .= "\t>> $up : Username-part include invalid charactor.\n";
	$e{'resp'} .= "\t>> subscribe failed.\n";
        return 0;
    }
    local(@ele);
    if ( $dom =~ /^\./ || $dom =~ /\.$/ || $dom =~ /\s/ ){
	$e{'resp'} .= "\t>> $addr : Domain-part include invalid dot notation or space.\n";
	$e{'resp'} .= "\t>> subscribe failed.\n";
	return 0;
    }
    @ele = split(/\./,$dom);
    foreach $atom (@ele){
	if ( $atom =~ /^([a-zA-Z0-9\-]+)$/ ){
	    # OK.
	}else{
	    $e{'resp'} .= "\t>> $addr : Domainpart include invalid charactor $atom\n";
	    $e{'resp'} .= "\t>> subscribe failed.\n";
	    return 0;
	}
    }
    return 1;
}


sub FormatAddr{
    $0 = "FormatAddr";
    my($addr) = @_;
    local($lp,$dp) = split(/\@/,$addr);
    $dp =~ tr/A-Z/a-z/;
    $addr = $lp.'@'.$dp;
    return $addr;
}

sub LcAddr{
    $0 = "LcAddr";
    my($addr) = @_;
    $addr =~ tr/A-Z/a-z/;
    return $addr;
}

sub PPLockWait{
    $0 = "PPLockWait";
    local($wait) = 0;
    unless(-d "$Savedir"){
	mkdir("$Savedir",0750);
	&Log("$0:FILE: \$Savedir($Savedir) is created.");
    }
    if ($USE_LOCKFILE == 1){
	while (-f "$Savedir/.lockfile"){
	    sleep 1;
	    $wait++;
	    if ($wait > 15){
		&Log("$0:Fail: Timeout to wait unlock ($Savedir/.lockfile)");
		print STDERR "$0:Fail: Timeout to wait unlock ($Savedir/.lockfile)\n";
		exit $EX_TEMPFAIL;
	    }
	}
	open (LOCK, "> $Savedir/$locktmp");
	print LOCK "$$\n";
	close (LOCK);
	while (!link("$Savedir/$locktmp", "$Savedir/.lockfile")){
	    sleep 1;
	    $wait++;
	    if ($wait > 32){
		&Log("$0:Fail: Timeout to link ($Savedir/.lockfile)");
		print STDERR "$0:Fail: Timeout to link ($Savedir/.lockfile)\n";
		exit $EX_TEMPFAIL;
	    }
	}
	&Log("$0:LOCK: lock Lockfile($Savedir/.lockfile) -- session start");
    }else{
	while (!open(LOCK,"+< config.ph")){
	    sleep 1;
	    $wait++;
	    if ($wait > 15){
		&Log("$0:Fail: Timeout to open (config.ph)");
		print STDERR "$0:Fail: Timeout to open (config.ph)\n";
		exit $EX_TEMPFAIL;
	    }
	}
	while (!flock(LOCK,$LOCK_EX|$LOCK_NB)){
	    sleep 1;
	    $wait++;
	    if ($wait > 32){
		&Log("$0:Fail: Timeout to lock (config.ph)");
		print STDERR "$0:Fail: Timeout to lock (config.ph)\n";
		exit $EX_TEMPFAIL;
	    }
	}
	&Log("$0:LOCK: lock config.ph -- session start");
    }
    return 1;
}

sub UnlockDone{
    $0 = "UnlockDone";
    &Log("$0:DONE: unlock -- session completed.");
    if ($USE_LOCKFILE == 1){
	unlink("$Savedir/.lockfile");
	unlink("$Savedir/$locktmp");
    }else{
	unless (flock(LOCK,$LOCK_UN)){
	    &Log("$0:Fail: Can't unlock (config.ph)") if $debug;
	}
	close(LOCK);
    }
}

sub UnlockDone2{
    $0 = "UnlockDone2";
    &Log("$0:DONE: unlock2 -- session aborted.");
    if ($USE_LOCKFILE == 1){
	unlink("$Savedir/.lockfile");
	unlink("$Savedir/$locktmp");
    }else{
	unless (flock(LOCK,$LOCK_UN)){
	    &Log("$0:Fail: Can't unlock (config.ph)") if $debug;
	}
	close(LOCK);
    }
    exit $EX_UNAVAILABLE;
}

sub DebugPrint{
    local(*e) = @_;
    foreach $key (sort(keys(%e))){
	print "-------- $key:\n$e{$key}\n";
    }
    return 1;
}

sub Log{
    local($s) = @_;
    ($sec,$min,$hour,$mday,$mon,$year,$wday) = (localtime(time))[0..6];
    $now = sprintf("%04d.%02d.%02d %02d:%02d:%02d",
		   1900 + $year, $mon + 1, $mday, $hour, $min ,$sec);
    $s =~ s/\r//;
    $s =~ s/\n//;
    open(LOG,">>$Logfile");
    print LOG "$now $s\n";
    close(LOG);
    return 1;
}


package ml;

sub main'SetML{ #'
    local($ml,*e)=@_;
    local($ML_SETTING_ROOT) = $main'ML_SETTING_ROOT; #'
    $ml =~ tr/A-Z/a-z/;
    local($DIR)="$ML_SETTING_ROOT/$ml";
    if ((-d $DIR) && (-x $DIR)){
	unshift(@INC, $DIR);
	chdir $DIR;
    }else{
	print STDERR "$0:WARN: \$DIR is not given or Can't chdir.";
	exit $EX_NOPERM;
    }
    &ReadConfig;
    &main'Log("$0:FROM: $e{'gecos'} $e{'poster'}"); #'
    ### limited value load to main NS
    foreach (keys(%USERCOMMPERM)){
        $main'USERCOMMPERM{$_} = $USERCOMMPERM{$_} || $main'USERCOMMPERM{$_};  #';
    }
    @main'DELIVERLIST = @DELIVERLIST; #'
    @main'MEMBERLIST = @MEMBERLIST; #'
    @main'ADMINLIST = @ADMINLIST; #'
    $main'ML_ADMIN = $ML_ADMIN; #'
    $main'ML_ADDR = $ML_ADDR; #'
    $main'BRACKET = $BRACKET; #'
    $main'AUTOSUBSCRIBE = $AUTOSUBSCRIBE; #'
    $main'OPTIMIZED_DELIVERLIST = $OPTIMIZED_DELIVERLIST; #'
    return 1;
}

sub ReadConfig{
    $0 = "ReadConfig";
    delete $INC{'config.ph'};
    eval("require(\"config.ph\")");
    if ($@) {
	&main'Log("$0:ERROR:$@");
	print STDERR $@;
	exit $EX_CONFIG;
    }
    return 1;
}
